#
#  engine
#
#  Playout Daemon for autoradio project
#
#
#  Copyright (C) 2017-2018 Gottfried Gaisbauer <gottfried.gaisbauer@servus.at>
#
#  This file is part of engine.
#
#  engine is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  engine is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with engine. If not, see <http://www.gnu.org/licenses/>.
#

import time
import logging
import json

from modules.communication.liquidsoap.playerclient import LiquidSoapPlayerClient
# from modules.communication.liquidsoap.recorderclient import LiquidSoapRecorderClient
from modules.communication.liquidsoap.initthread import LiquidSoapInitThread
from modules.communication.mail.mail import AuraMailer

from libraries.enum.auraenumerations import TerminalColors, ScheduleEntryType
from libraries.exceptions.auraexceptions import LQConnectionError
#from libraries.database.broadcasts import TrackService
from libraries.exceptions.exception_logger import ExceptionLogger

""" 
    LiquidSoapCommunicator Class 
    Uses LiquidSoapClient, but introduces more complex commands, transactions and error handling
"""

class LiquidSoapCommunicator(ExceptionLogger):
    client = None
    logger = None
    transaction = 0
    channels = None
    scheduler = None
    error_data = None
    auramailer = None
    is_liquidsoap_running = False
    connection_attempts = 0
    active_channel = None
    disable_logging = False
    fade_in_active = False
    fade_out_active = False

    # ------------------------------------------------------------------------------------------ #
    def __init__(self, config):
        """
        Constructor
        """
        self.config = config
        self.logger = logging.getLogger("AuraEngine")

        self.client = LiquidSoapPlayerClient(config, "engine.sock")
        # self.lqcr = LiquidSoapRecorderClient(config, "record.sock")

        errors_file = self.config.get("install_dir") + "/errormessages/controller_error.js"
        f = open(errors_file)
        self.error_data = json.load(f)
        f.close()

        self.auramailer = AuraMailer(self.config)

        self.is_liquidsoap_up_and_running()

    # ------------------------------------------------------------------------------------------ #
    def is_liquidsoap_up_and_running(self):
        try:
            self.uptime()
            self.is_liquidsoap_running = True
        except LQConnectionError as e:
            self.logger.info("Liquidsoap is not running so far")
            self.is_liquidsoap_running = False
        except Exception as e:
            self.logger.error("Cannot check if Liquidsoap is running. Reason: " + str(e))
            self.is_liquidsoap_running = False

    # ------------------------------------------------------------------------------------------ #
    def set_volume(self, mixernumber, volume):
        return self.__send_lqc_command__(self.client, "mixer", "volume", mixernumber, volume)

    # ------------------------------------------------------------------------------------------ #
    def get_active_mixer(self):
        """
        get active mixer in liquidsoap server
        :return:
        """
        activeinputs = []

        # enable more control over the connection
        self.enable_transaction()

        inputs = self.get_all_channels()

        cnt = 0
        for input in inputs:
            status = self.__get_mixer_status__(cnt)

            if "selected=true" in status:
                activeinputs.append(input)

            cnt = cnt + 1

        self.disable_transaction()

        return activeinputs

    # ------------------------------------------------------------------------------------------ #
    def get_active_channel(self):
        """
        Retrieves the active channel from programme.

        Returns:
            (String):   The channel type, empty string if no channel is active.
        """
        active_entry = self.scheduler.get_active_entry()
        if active_entry is None:
            return ""
        return active_entry.type

    # ------------------------------------------------------------------------------------------ #
    def get_mixer_status(self):
        inputstate = {}

        self.enable_transaction()

        inputs = self.get_all_channels()

        cnt = 0
        for input in inputs:
            inputstate[input] = self.__get_mixer_status__(cnt)
            cnt = cnt + 1

        self.disable_transaction()

        return inputstate

    # ------------------------------------------------------------------------------------------ #
    def get_mixer_volume(self, channel):
        return False

    # ------------------------------------------------------------------------------------------ #
    def get_recorder_status(self):
        self.enable_transaction(self.client)
        recorder_state = self.__send_lqc_command__(self.client, "record", "status")
        self.disable_transaction(self.client)

        return recorder_state

    # ------------------------------------------------------------------------------------------ #
    def http_start_stop(self, start):
        if start:
            cmd = "start"
        else:
            cmd = "stop"

        try:
            self.enable_transaction()
            self.__send_lqc_command__(self.client, "http", cmd)
            self.disable_transaction()
        except LQConnectionError:
            # we already caught and handled this error in __send_lqc_command__, but we do not want to execute this function further
            pass

    # ------------------------------------------------------------------------------------------ #
    def recorder_stop(self):
        self.enable_transaction()

        for i in range(5):
            if self.config.get("rec_" + str(i)) == "y":
                self.__send_lqc_command__(self.client, "recorder_" + str(i), "stop")

        self.disable_transaction()

    # ------------------------------------------------------------------------------------------ #
    def recorder_start(self, num=-1):
        if not self.is_liquidsoap_running:
            if num==-1:
                msg = "Want to start recorder, but LiquidSoap is not running"
            else:
                msg = "Want to start recorder " + str(num) + ", but LiquidSoap is not running"
            self.logger.warning(msg)
            return False

        self.enable_transaction()

        if num == -1:
            self.recorder_start_all()
        else:
            self.recorder_start_one(num)

        self.disable_transaction()

    # ------------------------------------------------------------------------------------------ #
    def recorder_start_all(self):
        if not self.is_liquidsoap_running:
            self.logger.warning("Want to start all recorder, but LiquidSoap is not running")
            return False

        self.enable_transaction()
        for i in range(5):
            self.recorder_start_one(i)
        self.disable_transaction()

    # ------------------------------------------------------------------------------------------ #
    def recorder_start_one(self, num):
        if not self.is_liquidsoap_running:
            return False

        if self.config.get("rec_" + str(num)) == "y":
            returnvalue = self.__send_lqc_command__(self.client, "recorder", str(num), "status")

            if returnvalue == "off":
                self.__send_lqc_command__(self.client, "recorder", str(num), "start")

    # ------------------------------------------------------------------------------------------ #
    def fade_in(self, new_entry):
        try:
            fade_in_time = float(self.config.get("fade_in_time"))

            if fade_in_time > 0:
                self.fade_in_active = True
                target_volume = new_entry.volume

                step = fade_in_time / target_volume

                self.logger.info("Starting to fading " + new_entry.type.value + " in. step is " + str(step) + "s. target volume is " + str(target_volume))

                self.disable_logging = True
                self.client.disable_logging = True

                for i in range(target_volume):
                    self.channel_volume(new_entry.type.value, i + 1)
                    time.sleep(step)

                self.logger.info("Finished with fading " + new_entry.type.value + " in.")

                self.fade_in_active = False
                if not self.fade_out_active:
                    self.disable_logging = False
                    self.client.disable_logging = False
        except LQConnectionError as e:
            self.logger.critical(str(e))

        return True

    # ------------------------------------------------------------------------------------------ #
    def fade_out(self, old_entry):
        try:
            fade_out_time = float(self.config.get("fade_out_time"))

            if fade_out_time > 0:
                step = abs(fade_out_time) / old_entry.volume

                self.logger.info("Starting to fading " + old_entry.type.value + " out. step is " + str(step) + "s")

                # disable logging... it is going to be enabled again after fadein and -out is finished
                self.disable_logging = True
                self.client.disable_logging = True

                for i in range(old_entry.volume):
                    self.channel_volume(old_entry.type.value, old_entry.volume-i-1)
                    time.sleep(step)

                self.logger.info("Finished with fading " + old_entry.type.value + " out.")

                # enable logging again
                self.fade_out_active = False
                if not self.fade_in_active:
                    self.disable_logging = False
                    self.client.disable_logging = False
        except LQConnectionError as e:
            self.logger.critical(str(e))

        return True

    # ------------------------------------------------------------------------------------------ #
    def activate(self, new_entry, cue_in=0.0):
        """
        Activates a new Playlist Entry.

        Args:
            new_entry (PlaylistEntry): The track to be played
        
        Raises:
            (LQConnectionError): In case connecting to LiquidSoap isn't possible
        """
        
        # Grab the actual active entry
        active_entry = self.scheduler.get_active_entry()
        # Set default channel, if no previous track is available
        current_channel = ScheduleEntryType.FILESYSTEM 
        if active_entry:
            current_channel = active_entry.type

        try:
            self.enable_transaction()
            if current_channel == new_entry.type:
                self.activate_same_channel(new_entry, cue_in)
            else:
                self.activate_different_channel(new_entry, cue_in, current_channel)
            self.disable_transaction()

            # FIXME Implement TrackService bi-directionally, log fallbacks too.
            self.logger.critical("FIXME: Implement TrackService")
#           self.insert_track_service_entry(new_entry)
        except LQConnectionError:
            # we already caught and handled this error in __send_lqc_command__, 
            # but we do not want to execute this function further and pass the exception
            pass


    def activate_same_channel(self, entry, cue_in=0.0, activate_different_channel=False):
        """
        Activates a playlist entry for the current channel.

        Args:
            entry (PlaylistEntry):  The entry to play.
            cue_in (Float):         A value in seconds where to cue the start of the entry.
        """
        if not activate_different_channel:
            self.logger.info(TerminalColors.PINK.value + entry.type.value + " already active!" + TerminalColors.ENDC.value)

        # Check if it needs to be pushed to a filesystem queue or stream
        if entry.type == ScheduleEntryType.FILESYSTEM:
            uri = entry.filename
            if cue_in > 0.0:
                uri = "annotate:liq_cue_in=\"%s\":%s" % (str(cue_in), entry.filename)
            self.playlist_push(uri)
            self.active_channel = entry.type

        elif entry.type == ScheduleEntryType.STREAM:
            self.set_http_url(entry.source)
            self.http_start_stop(True)
            self.active_channel = entry.type

        # else: # live
        # Nothing to do when we are live => just leave it as is

        self.active_channel = entry.type

        # Set active channel to wanted volume
        if not activate_different_channel:
            self.channel_volume(entry.type.value, entry.volume)


    def activate_different_channel(self, entry, cue_in, active_type):
        """
        Activates a playlist entry for a channel other then the currently active one.

        Args:
            entry (PlaylistEntry):            The entry to play.
            cue_in (Float):                   A value in seconds where to cue the start of the entry.
            active_type (ScheduleEntryType):  The type of the currently active channel
        """     
        self.logger.info(TerminalColors.PINK.value + "LiquidSoapCommunicator is activating " + entry.type.value + " & deactivating " + active_type.value + "!" + TerminalColors.ENDC.value)

        # Reuse of this function, because activate_same_channel and activate_different_channel 
        # are doing pretty the same except setting of the volume to zero
        self.activate_same_channel(entry, cue_in, True)

        # Set other channels to zero volume
        others = self.all_inputs_but(entry.getChannel())
        for o in others:
            self.channel_volume(o, 0)

        # Set active channel to wanted volume
        self.channel_volume(entry.type.value, entry.volume)

    # ------------------------------------------------------------------------------------------ #
    def insert_track_service_entry(self, schedule_entry):
        # create trackservice entry
        trackservice_entry = TrackService()

        # set foreign keys
        trackservice_entry.playlist_id = schedule_entry.playlist_id
        trackservice_entry.entry_num = schedule_entry.entry_num
        trackservice_entry.source = schedule_entry.source

        # store
        trackservice_entry.store(add=True, commit=True)

    # ------------------------------------------------------------------------------------------ #
    def all_inputs_but(self, input_type):
        try:
            activemixer_copy = self.get_all_channels().copy()
            activemixer_copy.remove(input_type)
        except ValueError as e:
            self.logger.error("Requested channel (" + input_type + ") not in channellist. Reason: " + str(e))
        except AttributeError:
            self.logger.critical("Channellist is None")

        return activemixer_copy

    # ------------------------------------------------------------------------------------------ #
    def get_all_channels(self):
        if self.channels is None or len(self.channels) == 0:
            self.channels = self.__send_lqc_command__(self.client, "mixer", "inputs")

        return self.channels

    # ------------------------------------------------------------------------------------------ #
    def reload_channels(self):
        self.channels = None
        return self.get_all_channels()

    # ------------------------------------------------------------------------------------------ #
    def __get_mixer_status__(self, mixernumber):
        return self.__send_lqc_command__(self.client, "mixer", "status", mixernumber)

    # ------------------------------------------------------------------------------------------ #
    def init_player(self):
        """
        Initializes the LiquidSoap Player after startup of the engine.

        Returns:
            (String):   Message that the player is started.
        """
        active_entry = self.scheduler.get_active_entry()

        t = LiquidSoapInitThread(self, active_entry)
        t.start()

        return "LiquidSoapInitThread started!"

    # ------------------------------------------------------------------------------------------ #
    def channel_activate(self, channel, activate):
        channels = self.get_all_channels()

        try:
            index = channels.index(channel)
            if len(channel) < 1:
                self.logger.critical("Cannot activate channel. There are no channels!")
            else:
                message = self.__send_lqc_command__(self.client, "mixer", "select", index, activate)
                return message
        except Exception as e:
            self.logger.critical("Ran into exception when activating channel. Reason: " + str(e))

    # ------------------------------------------------------------------------------------------ #
    def channel_volume(self, channel, volume):
        """
        set volume of a channel
        @type     channel:  string
        @param    channel:  Channel
        @type     volume:   int
        @param    volume:   Volume between 0 and 100
        """

        try:
            channels = self.get_all_channels()
            index = channels.index(channel)
        except ValueError as e:
            self.logger.error("Cannot set volume of channel " + channel + " to " + str(volume) + "!. Reason: " + str(e))
            return

        try:
            if len(channel) < 1:
                self.logger.warning("Cannot set volume of channel " + channel + " to " + str(volume) + "! There are no channels.")
            else:
                message = self.__send_lqc_command__(self.client, "mixer", "volume", str(index), str(int(volume)))

                if not self.disable_logging:
                    if message.find('volume=' + str(volume) + '%'):
                        self.logger.debug("Set volume of channel " + channel + " to " + str(volume))
                    else:
                        self.logger.warning("Setting volume of channel " + channel + " gone wrong! Liquidsoap message: " + message)

                return message
        except AttributeError as e: #(LQConnectionError, AttributeError):
            self.disable_transaction(force=True)
            self.logger.error("Ran into exception when setting volume of channel " + channel + ". Reason: " + str(e))

    # ------------------------------------------------------------------------------------------ #
    def auraengine_state(self):
        state = self.__send_lqc_command__(self.client, "auraengine", "state")
        return state

    # ------------------------------------------------------------------------------------------ #
    def liquidsoap_help(self):
        data = self.__send_lqc_command__(self.client, "help", "")
        if not data:
            self.logger.warning("Could not get Liquidsoap's help")
        else:
            self.logger.debug("Got Liquidsoap's help")
        return data

    # ------------------------------------------------------------------------------------------ #
    def set_http_url(self, uri):
        return self.__send_lqc_command__(self.client, "http", "url", uri)

    # ------------------------------------------------------------------------------------------ #
    def playlist_push(self, uri):
        """
        Adds an filesystem URI to the playlist

        Args:
            uri (String):   The URI of the file
        Returns:
            LiquidSoap Response
        """
        return self.__send_lqc_command__(self.client, "fs", "push", uri)

    # ------------------------------------------------------------------------------------------ #
    def playlist_seek(self, seconds_to_seek):
        """
        Forwards the player (n) seconds.

        Args:
            seconds_to_seeks (Float):   The seconds to skip
        """
        return self.__send_lqc_command__(self.client, "fs", "seek", str(seconds_to_seek))

    # ------------------------------------------------------------------------------------------ #
    def version(self):
        """
        get version
        """
        data = self.__send_lqc_command__(self.client, "version", "")
        self.logger.debug("Got Liquidsoap's version")
        return data

    # ------------------------------------------------------------------------------------------ #
    def uptime(self):
        """
        get uptime
        """
        data = self.__send_lqc_command__(self.client, "uptime", "")
        self.logger.debug("Got Liquidsoap's uptime")
        return data

    # ------------------------------------------------------------------------------------------ #
    def __send_lqc_command__(self, lqs_instance, namespace, command, *args):
        """
        Ein Kommando an Liquidsoap senden
        @type  lqs_instance: object
        @param lqs_instance: Instance of LiquidSoap Client
        @type  namespace:    string
        @param namespace:    Namespace of function
        @type  command:      string
        @param command:      Function name
        @type args:          list
        @param args:         List of parameters
        @rtype:              string
        @return:             Response from LiquidSoap
        """
        try:
            if not self.disable_logging:
                if namespace == "recorder":
                    self.logger.info("LiquidSoapCommunicator is calling " + str(namespace) + "_" + str(command) + "." + str(args))
                else:
                    if command == "":
                        self.logger.info("LiquidSoapCommunicator is calling " + str(namespace) + str(args))
                    else:
                        self.logger.info("LiquidSoapCommunicator is calling " + str(namespace) + "." + str(command) + str(args))

            # call wanted function ...
            func = getattr(lqs_instance, namespace)
            # ... and fetch the result
            result = func(command, *args)

            if not self.disable_logging:
                self.logger.debug("LiquidSoapCommunicator got response " + str(result))

            self.connection_attempts = 0

            return result

        except LQConnectionError as e:
            self.logger.error("Connection Error when sending " + str(namespace) + "." + str(command) + str(args))
            if self.try_to_reconnect():
                time.sleep(0.2)
                self.connection_attempts += 1
                if self.connection_attempts < 5:
                    # reconnect
                    self.__open_conn(self.client)
                    self.logger.info("Trying to resend " + str(namespace) + "." + str(command) + str(args))
                    # grab return value
                    retval = self.__send_lqc_command__(lqs_instance, namespace, command, *args)
                    # disconnect
                    self.__close_conn(self.client)
                    # return the val
                    return retval
                else:
                    if command == "":
                        msg = "Rethrowing Exception while trying to send " + str(namespace) + str(args)
                    else:
                        msg = "Rethrowing Exception while trying to send " + str(namespace) + "." + str(command) + str(args)

                    self.logger.info(msg)
                    self.disable_transaction(socket=self.client, force=True)
                    raise e
            else:
                # also store when was last admin mail sent with which content...
                # FIXME implement admin mail sending
                self.logger.critical("SEND ADMIN MAIL AT THIS POINT")
                raise e

    # ------------------------------------------------------------------------------------------ #
    def try_to_reconnect(self):
        self.enable_transaction()
        return self.transaction > 0

    # ------------------------------------------------------------------------------------------ #
    def enable_transaction(self, socket=None):
        # set socket to playout if nothing else is given
        if socket is None:
            socket = self.client

        self.transaction = self.transaction + 1

        self.logger.debug(TerminalColors.WARNING.value + "ENabling transaction! cnt: " + str(self.transaction) + TerminalColors.ENDC.value)

        if self.transaction > 1:
            return

        try:
            self.__open_conn(socket)
        except FileNotFoundError:
            self.disable_transaction(socket=socket, force=True)

            msg = "socket file " + socket.socket_path + " not found. Is liquidsoap running?"
            self.logger.critical(TerminalColors.RED.value + msg + TerminalColors.ENDC.value)
            self.auramailer.send_admin_mail("CRITICAL Exception when connecting to Liquidsoap", msg)

    # ------------------------------------------------------------------------------------------ #
    def disable_transaction(self, socket=None, force=False):
        if not force:
            # nothing to disable
            if self.transaction == 0:
                return

            # decrease transaction counter
            self.transaction = self.transaction - 1

            # debug msg
            self.logger.debug(TerminalColors.WARNING.value + "DISabling transaction! cnt: " + str(self.transaction) + TerminalColors.ENDC.value)

            # return if connection is still needed
            if self.transaction > 0:
                return
        else:
            self.logger.debug(TerminalColors.WARNING.value + "Forcefully DISabling transaction! " + TerminalColors.ENDC.value)

        # close conn and set transactioncounter to 0
        self.__close_conn(socket)
        self.transaction = 0

    # ------------------------------------------------------------------------------------------ #
    def __open_conn(self, socket):
        # already connected
        if self.transaction > 1:
            return

        self.logger.debug(TerminalColors.GREEN.value + "LiquidSoapCommunicator opening conn" + TerminalColors.ENDC.value)

        # try to connect
        socket.connect()

    # ------------------------------------------------------------------------------------------ #
    def __close_conn(self, socket):
        # set socket to playout
        if socket is None:
            socket = self.client

        # do not disconnect if a transaction is going on
        if self.transaction > 0:
            return

        # say bye
        socket.byebye()

        # debug msg
        self.logger.debug(TerminalColors.BLUE.value + "LiquidSoapCommunicator closed conn" + TerminalColors.ENDC.value)
